"""
iTunes Smart playlist parser with Python. Decode itunes smart playlist rules from base64 data. Convert iTunes smart playlists to Kodi xsp smart playlists.
"""

__version__ = '1.1.5'
__author__ = 'cuzi'
__email__ = 'cuzi@openmail.cc'
__source__ = 'https://github.com/cvzi/itunes_smartplaylist'
__license__ = """
MIT License

Copyright (c) cuzi 2018

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__all__ = ["Parser", "SmartPlaylist", "BytesParser", "createXSPFile", "createXSP", "PlaylistException", "EmptyPlaylistException", "readiTunesLibrary", "generatePersistentIDMapping", "createPlaylistTree", "LibraryException"]

from itunessmart.parse import SmartPlaylistParser, SmartPlaylist
from itunessmart.xsp import createXSPFile, createXSP, PlaylistException, EmptyPlaylistException
from itunessmart.library import readiTunesLibrary, generatePersistentIDMapping, createPlaylistTree, LibraryException


class Parser:
    """Parse data from a base64 encoded string"""
    def __init__(self, datastr_info: str = None, datastr_criteria: str = None):
        """Parse data from a base64 encoded string"""
        self.result = None
        self._parser = SmartPlaylistParser(datastr_info, datastr_criteria)
        if datastr_info and datastr_criteria:
            self._update()

    def _update(self):
        self.result = None
        self._parser.parse()
        self.result = self._parser.result()

    def update_data_base64(self, datastr_info: str, datastr_criteria: str) -> SmartPlaylist:
        self._parser.str_data(datastr_info, datastr_criteria)
        self._update()
        return self.result

    def update_data_bytes(self, data_info: bytes, data_criteria: bytes) -> SmartPlaylist:
        self._parser.data(data_info, data_criteria)
        self._update()
        return self.result


class BytesParser(Parser):
    """Parse data from raw bytes"""
    def __init__(self, data_info, data_criteria):
        """Parse data from raw bytes"""
        super().__init__()
        self._parser = SmartPlaylistParser()
        self._parser.data(data_info, data_criteria)
        self._parser.parse()

        self._update()



"""
Module holding static data related to the iTunes playlist format.
"""

from enum import IntEnum


class FileKind:
    def __init__(self, name, extension):
        self.name = name
        self.extension = extension


FileKinds = [
    FileKind("Protected AAC audio file", ".m4p"),
    FileKind("MPEG audio file", ".mp3"),
    FileKind("AIFF audio file", ".aiff"),
    FileKind("WAV audio file", ".wav"),
    FileKind("QuickTime movie file", ".mov"),
    FileKind("MPEG-4 video file", ".mp4"),
    FileKind("AAC audio file", ".m4a")
]

MediaKinds = {
        0x01: "Music",
        0x02: "Movie",
        0x04: "Podcast",
        0x08: "Audiobook",
        0x20: "Music Video",
        0x40: "TV Show",
       0x400: "Home Video",
     0x10000: "iTunes Extras",
    0x100000: "Voice Memo",
    0x200000: "iTunes U",
    0xC00000: "Book",
    # TODO mediakinds of toplevel playlists
    # The following are only used in the toplevel playlists: Music, TV Shows, Movies and Books
    # (These playlists can be selected with the arrows or from a dropdown menu in iTunes 12)
    # Theses Media Kinds cannot be selected by the user. I am unsure about their meaning.
    0xC00008: "Book or Audiobook",  # TODO My guess: This contains Books and Audiobooks
    0x1021B1: "Music",  # TODO My guess: This is similar to Music
    0x208004: "Undesired Music",  # TODO My guess: This is some kind of Music that should not appear in the toplevel playlist
    0x20A004: "Undesired Other"  # TODO My guess: This is something (other than music) that should not appear in the toplevel playlist
}

iCloudStatus = {
    0x01: "Purchased",
    0x02: "Matched",
    0x03: "Uploaded",
    0x04: "Ineligible",
    0x05: "Local Only",
    0x07: "Duplicate"
}

LoveStatus = {
    0x00: "None",
    0x02: "Loved",
    0x03: "Disliked",
}

LocationKinds = {
    0x01: "Computer",
    0x10: "iCloud"
}

SelectionMethodsStrings = {
    "Random": ["random", "RANDOM()"],
    "Name": ["name", "SortName"],
    "Album": ["album", "SortAlbum"],
    "Artist": ["artist", "SortArtist"],
    "Genre": ["genre", "Genre"],
    "HighestRating": ["highest rated", "Rating DESC"],
    "LowestRating": ["lowest rated", "Rating ASC"],
    "RecentlyPlayed": [["most recently played", "least recently played"], ["LastPlayed DESC", "LastPlayed ASC"]],
    "OftenPlayed": [["most often played", "least often played"], ["Plays DESC", "Plays ASC"]],
    "RecentlyAdded": [["most recently added", "least recently added"], ["DateAdded DESC", "DateAdded ASC"]]
}


DateStartFromUnix = -2082844800  # iTunes/Unix time stamp 0 difference


class LimitMethods(IntEnum):
    """The methods by which the number of songs in a playlist are limited"""
    Minutes = 0x01
    MB = 0x02
    Items = 0x03
    Hours = 0x04
    GB = 0x05


class SelectionMethods(IntEnum):
    """The methods by which songs are selected for inclusion in a limited playlist"""
    Random = 0x02
    Name = 0x05
    Album = 0x06
    Artist = 0x07
    Genre = 0x09
    HighestRating = 0x1c
    LowestRating = 0x01
    RecentlyPlayed = 0x1a
    OftenPlayed = 0x19
    RecentlyAdded = 0x15


class StringFields(IntEnum):
    """The matching criteria which take string data"""
    Album = 0x03
    AlbumArtist = 0x47
    Artist = 0x04
    Category = 0x37
    Comments = 0x0e
    Composer = 0x12
    Description = 0x36
    Genre = 0x08
    Grouping = 0x27
    Kind = 0x09
    Name = 0x02
    Show = 0x3e
    SortAlbum = 0x4f
    SortAlbumartist = 0x51
    SortComposer = 0x52
    SortName = 0x4e
    SortShow = 0x53
    VideoRating = 0x59


class IntFields(IntEnum):
    """The matching criteria which take integer data"""
    BPM = 0x23
    BitRate = 0x05
    Compilation = 0x1f
    DiskNumber = 0x18
    Plays = 0x16
    Rating = 0x19
    Podcast = 0x39
    SampleRate = 0x06
    Season = 0x3f
    Size = 0x0c
    Skips = 0x44
    Duration = 0x0d
    TrackNumber = 0x0b
    Year = 0x07


class BooleanFields(IntEnum):
    """The matching criteria which take boolean data"""
    HasArtwork = 0x25
    Purchased = 0x29
    Checked = 0x1d


class DateFields(IntEnum):
    """The matching criteria which take date data"""
    DateAdded = 0x10
    DateModified = 0x0a
    LastPlayed = 0x17
    LastSkipped = 0x45


class MediaKindFields(IntEnum):
    """The matching criteria which take a Media Kind, as defined above"""
    MediaKind = 0x3c


class PlaylistFields(IntEnum):
    """The matching criteria which take a Persistent Playlist ID (64bit)"""
    PlaylistPersistentID = 0x28


class LoveFields(IntEnum):
    """The matching criteria which take a Persistent Playlist ID (64bit)"""
    Love = 0x9a


class CloudFields(IntEnum):
    """The matching criteria which take iCloud status , as defined above """
    iCloudStatus = 0x86


class LocationFields(IntEnum):
    """The matching criteria which take a Location, as defined above"""
    Location = 0x85


class LogicSign(IntEnum):
    """The signs which apply to different kinds of logic (is vs. is not, contains vs. doesn't contain, etc.)"""
    IntPositive = 0x00
    StringPositive = 0x01
    IntNegative = 0x02
    StringNegative = 0x03


class LogicRule(IntEnum):
    """The logical rules"""
    Other = 0x00
    Is = 0x01
    Contains = 0x02
    Starts = 0x04
    Ends = 0x08
    Greater = 0x10
    Less = 0x40


class Offset(IntEnum):
    """Byte offsets for the fields"""
    INTLENGTH = 67             # The length on a int criteria starting at the first int
    SUBEXPRESSIONLENGTH = 192  # The length of a subexpression starting from FIELD

    # INFO OFFSETS
    # Offsets for bytes which...
    LIVEUPDATE = 0            # determine whether live updating is enabled - Absolute offset
    MATCHBOOL = 1             # determine whether logical matching is to be performed - Absolute offset
    LIMITBOOL = 2             # determine whether results are limited - Absolute offset
    LIMITMETHOD = 3           # determine by what criteria the results are limited - Absolute offset
    SELECTIONMETHOD = 7       # determine by what criteria limited playlists are populated - Absolute offset
    LIMITINT = 8              # determine the limited - Absolute offset
    LIMITCHECKED = 12         # determine whether to exclude unchecked items - Absolute offset
    SELECTIONMETHODSIGN = 13  # determine whether certain selection methods are "most" or "least" - Absolute offset

    # CRITERIA OFFSETS
    # Offsets for bytes which...
    LOGICTYPE = 15     # determine whether all or any criteria must match - Absolute offset
    FIELD = 139        # determine what is being matched (Artist, Album, &c) - Absolute offset
    LOGICSIGN = 1      # determine whether the matching rule is positive or negative (e.g., is vs. is not) - Relative offset from FIELD
    LOGICRULE = 4      # determine the kind of logic used (is, contains, begins, &c) - Relative offset from FIELD
    STRING = 54        # begin string data - Relative offset from FIELD
    INTA = 57          # begin the first int - Relative offset from FIELD
    INTB = 24          # begin the second int - Relative offset from INTA
    TIMEMULTIPLE = 73  # begin the int with the multiple of time - Relative offset from FIELD
    TIMEVALUE = 65     # begin the inverse int with the value of time - Relative offset from FIELD
    SUBLOGICTYPE = 68  # determine whether all or any criteria must match - Relative offset from FIELD
    SUBINT = 61        # begin the first int - Relative offset from FIELD



"""
Module holding static data related to the XSP playlist format.
"""

__all__ = [
    "xml_dec",
    "xml_doc",
    "xml_rule",
    "xml_value",
    "xsp_fields",
    "xsp_allowed_fields",
    "xsp_operators",
    "xsp_allowed_operators",
    "xsp_sorting"]

xml_dec = '<?xml version="1.0" encoding="UTF-8" standalone="yes" ?>'
xml_doc = '''{dec}
<smartplaylist type="songs">
    <name>{name}</name>
    <match>{globalmatch}</match>
{rules}
{meta}
</smartplaylist>'''

xml_rule = '''    <rule field="{field}" operator="{operator}">
{values}
    </rule>'''

xml_value = '''        <value>{value}</value>'''

xsp_fields = {
    "Artist": "artist",
    "AlbumArtist": "albumartist",
    "Album": "album",
    "Genre": "genre",
    "Name": "title",
    "Year": "year",
    "Duration": "time",
    "TrackNumber": "tracknumber",
    "Plays": "playcount",
    "LastPlayed": "lastplayed",
    "Rating": "userrating",
    "Comments": "comment",
    "PlaylistPersistentID": "playlist"
}

xsp_allowed_fields = xsp_fields.keys()

xsp_operators = {
    "and": "all",
    "or": "one",
    "like": "contains",
    "not like": "doesnotcontain",
    "is": "is",
    "is not": "isnot",
    "starts with": "startswith",
    "ends with": "endswith",
    "less than": "lessthan",
    "greater than": "greaterthan",
    "is after": "after",
    "is before": "before",
    "is in the last": "inthelast",
    "is not in the last": "notinthelast"
}

xsp_allowed_operators = xsp_operators.keys()

xsp_sorting = {
    "SortName": ("title", "ascending"),
    "SortAlbum": ("album", "ascending"),
    "SortArtist": ("artist", "ascending"),
    "Genre": ("genre", "ascending"),
    "Rating DESC": ("userrating", "descending"),
    "Rating ASC": ("userrating", "ascending"),
    "LastPlayed DESC": ("lastplayed", "descending"),
    "LastPlayed ASC": ("lastplayed", "ascending"),
    "Plays DESC": ("playcount", "descending"),
    "Plays ASC": ("playcount", "ascending"),
    "DateAdded DESC": ("dateadded ", "descending"),
    "DateAdded ASC": ("dateadded ", "ascending")
}



"""
Module for iTunes Library and library file `iTunes Music Library.xml`
"""

import time
import datetime
import base64
from typing import BinaryIO, Tuple, Dict

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET


class Library(dict):
    pass


class LibraryException(Exception):
    pass


def generatePersistentIDMapping(library: Library) -> Dict[str, str]:
    """Create a mapping from playlist id to playlist name. Necessary for converting rules concerning other playlists to xsp.
    :param dict library: the result of readiTunesLibrary()
    :return: persistentIDMapping
    :rtype: dict
    """
    persistentIDMapping = {}
    for playlist in library['Playlists']:
        if 'Playlist Persistent ID' in playlist and "Name" in playlist:
            persistentIDMapping[playlist['Playlist Persistent ID']] = playlist["Name"]
    return persistentIDMapping


def readiTunesLibrary(libraryFileStream: BinaryIO) -> Library:
    """Read itunes library file `iTunes Music Library.xml` and return dict
    :param stream libraryFileStream: file `iTunes Music Library.xml`
    :return: iTunes library content
    :rtype: Library
    """
    parser = ET.iterparse(libraryFileStream, events=('start', 'end'))
    _, plist = next(parser)

    if plist.tag != "plist":
        raise LibraryException("Root element is not <plist> element")
    if plist.attrib['version'] != "1.0":
        raise LibraryException("<plist> version is not 1.0")

    current = Library()
    current_islist = False  # current can be dict or list
    key = "plist"
    data = current
    parent = [data]

    for event, elem in parser:
        if event == "start":
            if elem.tag == 'dict':
                if current_islist:
                    t = {}
                    current.append(t)
                    parent.append(current)
                    current = t
                else:
                    current[key] = {}
                    parent.append(current)
                    current = current[key]
                current_islist = False
            elif elem.tag == 'key':
                key = elem.text
            elif elem.tag == 'array':
                if current_islist:
                    t = []
                    current.append(t)
                    parent.append(current)
                    current = t
                else:
                    current[key] = []
                    parent.append(current)
                    current = current[key]
                current_islist = True
            # else:
            #     pass
        elif event == "end":
            # elif elem.tag == 'key':
            #     pass
            if elem.tag == 'dict' or elem.tag == 'array':
                current = parent.pop()
                current_islist = isinstance(current, list)
            else:
                if elem.tag == 'true':
                    elem.text = True
                elif elem.tag == 'false':
                    elem.text = False
                elif elem.tag == 'integer':
                    elem.text = int(elem.text)
                elif elem.tag == 'date':
                    try:
                        elem.text = int(time.mktime(datetime.datetime.strptime(elem.text, "%Y-%m-%dT%H:%M:%SZ").timetuple()))
                    except ValueError:
                        elem.text = 0
                    except OverflowError as e:
                        t = datetime.datetime.strptime(elem.text, "%Y-%m-%dT%H:%M:%SZ").timetuple()
                        if t.tm_year < 1971:
                            d = 1980 - t.tm_year
                            t2 = time.struct_time([t.tm_year + d, t.tm_mon, t.tm_mday, t.tm_hour, t.tm_min, t.tm_sec, t.tm_wday, t.tm_yday, t.tm_isdst])
                            elem.text = int(time.mktime(t2)) - d * 31557600
                        elif t.tm_year > 2030:
                            d = t.tm_year - 2040
                            t2 = time.struct_time([t.tm_year - d, t.tm_mon, t.tm_mday, t.tm_hour, t.tm_min, t.tm_sec, t.tm_wday, t.tm_yday, t.tm_isdst])
                            elem.text = int(time.mktime(t2)) + d * 31557600
                        else:
                            raise e

                elif elem.tag == 'data':
                    elem.text = base64.standard_b64decode("".join(elem.text.split()))

                if current_islist:
                    current.append(elem.text)
                else:
                    if elem.tag == 'key':
                        key = elem.text
                    elif key is not None:
                        current[key] = elem.text
                    else:
                        raise LibraryException("unexpected end tag %r" % (elem, ))

            elem.clear()
    plist.clear()

    data = data['plist']
    return data


class Node:
    def __init__(self, data):
        if isinstance(data, str):
            data = {'Name': data}
        self.data = data
        self.children = []
        self.parent = None

    def __str__(self):
        return "%s" % (str(self.data['Name']) if 'Name' in self.data else "Node:Unkown name")

    def __repr__(self):
        return "%s #%s" % (str(self.data['Name']) if 'Name' in self.data else "Node:Unkown name", str(self.data['Playlist Persistent ID']) if 'Playlist Persistent ID' in self.data else "")


def createPlaylistTree(library: Library) -> Tuple[Node, dict]:
    """ Create playlist tree
    :param Library library: the result of readiTunesLibrary()
    :return: Return the tree and a mapping from PersistentId to playlist: (rootNode, playlistByPersistentId_dict)
    :rtype: tuple
    """

    nodesByPersistentId = {}
    playlistByPersistentId = {}  # Map PlaylistPersistentId to Playlist data
    otherPlaylists = []  # Playlists without PersistendId
    childPlaylists = []  # This is a list of playlists, that have a Parent Persistent ID

    for playlist in library['Playlists']:
        # Clean up tracks array
        if 'Playlist Items' in playlist:
            playlist['Playlist Items'] = [[dictionary[x] for x in dictionary][0] for dictionary in playlist['Playlist Items']]

        if 'Playlist Persistent ID' not in playlist:
            otherPlaylists.append(playlist)
            continue
        playlistByPersistentId[playlist['Playlist Persistent ID']] = playlist
        if "Parent Persistent ID" in playlist:
            childPlaylists.append(playlist)

    for playlist in childPlaylists:
        node = Node(playlist)
        nodesByPersistentId[playlist['Playlist Persistent ID']] = node

        if playlist["Parent Persistent ID"] in nodesByPersistentId:
            parentNode = nodesByPersistentId[playlist["Parent Persistent ID"]]
        else:
            parentNode = Node(playlistByPersistentId[playlist["Parent Persistent ID"]])
            nodesByPersistentId[parentNode.data['Playlist Persistent ID']] = parentNode

        node.parent = parentNode
        parentNode.children.append(node)

    parentNodes = [(nodesByPersistentId[PerId] if PerId in nodesByPersistentId else Node(playlistByPersistentId[PerId])) for PerId in playlistByPersistentId if not playlistByPersistentId[PerId] in childPlaylists]

    root = Node("root")
    root.children = parentNodes + [Node(playlist) for playlist in otherPlaylists]
    root.children.sort(key=lambda node: node.data['Name'] if "Name" in node.data else "")

    return root, playlistByPersistentId



#! /usr/bin/env python3

"""Export smart playlists from "iTunes Music Library.xml" to xsp files for Kodi"""

__all__ = ["main"]

import os
import traceback
import sys

# If we are running from a wheel, add the wheel to sys.path
if __package__ == '':
    path = os.path.dirname(os.path.dirname(__file__))
    sys.path.insert(0, path)

try:
    import itunessmart
except ImportError:
    include = os.path.relpath(os.path.join(os.path.dirname(__file__), ".."))
    sys.path.insert(0, include)
    import itunessmart


def printWithoutException(s):
    try:
        print(s)
    except UnicodeEncodeError:
        try:
            print(str(s).encode(sys.stdout.encoding, errors='replace').decode(sys.stdout.encoding))
        except UnicodeEncodeError:
            print(str(s).encode('ascii', errors='replace').decode('ascii'))


def main(iTunesLibraryFile=None, outputDirectory=None):

    if iTunesLibraryFile == "--help" or iTunesLibraryFile == "-h" or iTunesLibraryFile == "/?":
        print("""Export smart playlists from 'iTunes Music Library.xml' to .xsp files for Kodi.
This interactive script uses the default iTunes library, but you may also specify a xml file:

Optional arguments:

python3 -m itunessmart {iTunesLibraryFile} {outputDirectory}

{iTunesLibraryFile}\t - Default is ~\\Music\\iTunes\\iTunes Music Library.xml
{outputDirectory}\t - Default is ./out/""")
        return

    export_sub_playlists = True

    if not iTunesLibraryFile:
        home = os.path.expanduser("~")
        folder = os.path.join(home, "Music/iTunes")
        iTunesLibraryFile = os.path.join(folder, "iTunes Music Library.xml")
    elif not os.path.isfile(iTunesLibraryFile) and os.path.isfile(os.path.join(iTunesLibraryFile, "iTunes Music Library.xml")):
        iTunesLibraryFile = os.path.join(iTunesLibraryFile, "iTunes Music Library.xml")

    while not os.path.isfile(iTunesLibraryFile):
        iTunesLibraryFile = input("# Please enter the path of your `iTunes Music Library.xml`: ")
        if os.path.isfile(iTunesLibraryFile):
            break
        elif os.path.isfile(os.path.join(iTunesLibraryFile, "iTunes Music Library.xml")):
            iTunesLibraryFile = os.path.join(iTunesLibraryFile, "iTunes Music Library.xml")
            break
        else:
            print("! Could not find file `%s`")

    print("# Reading %s . . . " % iTunesLibraryFile)
    with open(iTunesLibraryFile, "rb") as fs:
        # Read XML file
        library = itunessmart.readiTunesLibrary(fs)
        persistentIDMapping = itunessmart.generatePersistentIDMapping(library)

    print("# Library loaded!")

    userinput = input("# Do you want to convert a (single) or (all) playlists? ")
    export_all = True
    if userinput.lower() in ("single", "1", "one"):
        export_all = False

    userinput = input("# Do you want to export nested rules to sub-playlists? (yes/no) ")
    if userinput.lower() in ("n", "no", "0"):
        export_sub_playlists = False

    # Decode and export all smart playlists

    parser = itunessmart.Parser()

    if outputDirectory is None:
        outputDirectory = os.path.abspath("out")

    if not os.path.exists(outputDirectory):
        os.makedirs(outputDirectory)

    if export_all:
        print("# Converting playlists to %s" % outputDirectory)

    res = []
    for playlist in library['Playlists']:
        if 'Name' in playlist and 'Smart Criteria' in playlist and 'Smart Info' in playlist and playlist['Smart Criteria']:
            try:
                parser.update_data_bytes(playlist['Smart Info'], playlist['Smart Criteria'])
            except Exception as e:
                print("! Failed to decode playlist:")
                try:
                    print(traceback.format_exc())
                    printWithoutException(playlist['Name'])
                except KeyError:
                    printWithoutException(playlist)

            if not parser.result:
                continue

            try:
                res.append((playlist['Name'], parser.result))
                if export_all:
                    itunessmart.createXSPFile(directory=outputDirectory, name=playlist['Name'], smartPlaylist=parser.result, createSubplaylists=export_sub_playlists, persistentIDMapping=persistentIDMapping)
            except itunessmart.EmptyPlaylistException as e:
                printWithoutException("! `%s` is empty." % playlist['Name'])
            except itunessmart.PlaylistException as e:
                printWithoutException("! Skipped `%s`: %s" % (playlist['Name'], str(e)))
            except Exception as e:
                print("! Failed to convert playlist:")
                try:
                    print(traceback.format_exc())
                    printWithoutException(playlist['Name'])
                except KeyError:
                    printWithoutException(playlist)

    if not export_all:
        i = 1
        for p_name, _ in res:
            printWithoutException("# %02d: %s" % (i, p_name))
            i += 1

        while True:
            i = input("# Please give the number of the playlist (0 to exit): ")
            if i in ("0", ""):
                break

            i = int(i)
            p_name, p_result = res[i-1]
            printWithoutException("# Converting Playlist: %s" % p_name)
            try:
                itunessmart.createXSPFile(directory=outputDirectory, name=p_name, smartPlaylist=p_result, createSubplaylists=export_sub_playlists, persistentIDMapping=persistentIDMapping)
            except itunessmart.EmptyPlaylistException as e:
                printWithoutException("! `%s` is empty." % p_name)
            except itunessmart.PlaylistException as e:
                printWithoutException("! Skipped `%s`: %s" % (p_name, str(e)))
            except Exception as e:
                try:
                    print("! Failed to convert playlist:")
                    print(traceback.format_exc())
                    printWithoutException(p_name)
                except (UnicodeEncodeError, KeyError, TypeError) as e:
                    print(p_result)

    print("# All done!")
    userdata = os.path.expandvars(r"%appdata%\kodi\userdata\playlists\music") if os.path.exists(os.path.expandvars(r"%appdata%\kodi\userdata")) else os.path.expanduser(r"~/Library/Application Support/Kodi/userdata/playlists/music")
    print("# You may copy the .xsp files to %s" % userdata)


if __name__ == "__main__":
    main(*sys.argv[1:])



"""
Module holding the parsing algorithm
"""

import logging
import base64
import datetime
import struct
import json
from itunessmart.data_structure import *


class SmartPlaylist:
    """ Parser result. Contains all decoded playlist data"""

    def __init__(self, parser):
        self.output = parser.output
        self.query = parser.query
        self.queryTree = parser.queryTree
        self.ignore = parser.ignore

    def __str__(self):
        return "SmartPlaylist(`%s`)" % self.query

    def __repr__(self):
        return "SmartPlaylist(%s)" % json.dumps({"queryTree": self.queryTree, "ignore": self.ignore}, indent=2)


class SmartPlaylistParser:
    def __init__(self, datastr_info=None, datastr_criteria=None):
        self.is_parsed = False
        if datastr_info and datastr_criteria:
            self.str_data(datastr_info, datastr_criteria)
            self.parse()

    def str_data(self, datastr_info, datastr_criteria):
        i = base64.standard_b64decode("".join(datastr_info.split()))
        c = base64.standard_b64decode("".join(datastr_criteria.split()))
        return self.data(i, c)

    def data(self, data_info, data_criteria):
        self.is_parsed = False
        self.info = data_info
        self.criteria = data_criteria
        self.query = ""
        self.root = {}
        self.queryTree = self.root
        self.queryTreeCurrent = []
        self.fullTreeRoot = {}
        self.fullTree = self.fullTreeRoot
        self.fullTreeCurrent = []
        self.output = ""
        self.ignore = ""
        self.limit = {}

        self.subStack = []

    def result(self):
        if self.is_parsed:
            return SmartPlaylist(self)
        raise RuntimeError("Data not parsed yet. Call parse() before result()")

    def parse(self):
        if self.is_parsed:
            return

        if not hasattr(self, 'info') or not hasattr(self, 'criteria') or not self.info or not self.criteria:  # pragma: no cover
            raise RuntimeError("Set smart info with data() or strdata() before running parse()")

        self.offset = int(Offset.FIELD)

        if self.info[Offset.MATCHBOOL] == 1:
            self.is_or = self.criteria[Offset.LOGICTYPE] == 1
            if self.is_or:
                self.conjunctionQuery = " OR "
                self.conjunctionOutput = ' or\n'
                self.queryTree["or"] = self.queryTreeCurrent
                self.fullTree["or"] = self.fullTreeCurrent
            else:
                self.conjunctionQuery = " AND "
                self.conjunctionOutput = ' and\n'
                self.queryTree["and"] = self.queryTreeCurrent
                self.fullTree["and"] = self.fullTreeCurrent

            while True:
                self.again = False

                if len(self.subStack) > 0:
                    if self.subStack[-1]["N"] == 0:
                        old = self.subStack.pop()

                        self.query = old["query"] + \
                            old["conjunctionQuery"] + "( %s )" % self.query
                        self.output = old["output"] + old["conjunctionOutput"] + \
                            "[\n\t%s\n]" % "\n\t".join(self.output.split("\n"))

                        self.conjunctionQuery = old["conjunctionQuery"]
                        self.conjunctionOutput = old["conjunctionOutput"]
                        self.queryTreeCurrent = old["queryTreeCurrent"]
                        self.fullTreeCurrent = old["fullTreeCurrent"]
                    else:
                        self.subStack[-1]["N"] -= 1

                self.logicSignOffset = self.offset + int(Offset.LOGICSIGN)
                self.logicRulesOffset = self.offset + int(Offset.LOGICRULE)
                self.stringOffset = self.offset + int(Offset.STRING)
                self.intAOffset = self.offset + int(Offset.INTA)
                self.intBOffset = self.intAOffset + int(Offset.INTB)
                self.timeMultipleOffset = self.offset + \
                    int(Offset.TIMEMULTIPLE)
                self.timeValueOffset = self.offset + int(Offset.TIMEVALUE)

                if any(self.criteria[self.offset] ==
                       e.value for e in StringFields):
                    self.ProcessStringField()
                elif any(self.criteria[self.offset] == e.value for e in IntFields):
                    self.ProcessIntField()
                elif any(self.criteria[self.offset] == e.value for e in DateFields):
                    self.ProcessDateField()
                elif any(self.criteria[self.offset] == e.value for e in BooleanFields):
                    self.ProcessBooleanField()
                elif any(self.criteria[self.offset] == e.value for e in MediaKindFields):
                    self.ProcessListField(MediaKindFields, MediaKinds, listtype="mediakind")
                elif any(self.criteria[self.offset] == e.value for e in PlaylistFields):
                    self.ProcessPlaylistField()
                elif any(self.criteria[self.offset] == e.value for e in CloudFields):
                    self.ProcessListField(CloudFields, iCloudStatus, listtype="cloud")
                elif any(self.criteria[self.offset] == e.value for e in LoveFields):
                    self.ProcessListField(LoveFields, LoveStatus, listtype="love")
                elif any(self.criteria[self.offset] == e.value for e in LocationFields):
                    self.ProcessListField(LocationFields, LocationKinds, listtype="location")
                elif self.criteria[self.offset] == 0:
                    # Subexpression

                    self.is_or = self.criteria[self.offset +
                                               Offset.SUBLOGICTYPE] == 1

                    numberOfSubExpression = self._iTunesUint(
                        self.criteria[self.offset + Offset.SUBINT:self.offset + +Offset.SUBINT + 4])

                    self.subStack.append({
                        "query": self.query,
                        "output": self.output,
                        "N": numberOfSubExpression,
                        "conjunctionQuery": self.conjunctionQuery,
                        "conjunctionOutput": self.conjunctionOutput,
                        "queryTreeCurrent": self.queryTreeCurrent,
                        "fullTreeCurrent": self.fullTreeCurrent,
                    })

                    newtree = {}
                    newcurrent = []
                    self.queryTreeCurrent.append(newtree)

                    newfulltree = {}
                    newfulltreecurrent = []
                    self.fullTreeCurrent.append(newfulltree)

                    if self.is_or:
                        self.conjunctionQuery = " OR "
                        self.conjunctionOutput = ' or\n'
                        newtree["or"] = newcurrent
                        newfulltree["or"] = newfulltreecurrent
                    else:
                        self.conjunctionQuery = " AND "
                        self.conjunctionOutput = ' and\n'
                        newtree["and"] = newcurrent
                        newfulltree["and"] = newfulltreecurrent

                    self.query = ""
                    self.output = ""
                    self.queryTreeCurrent = newcurrent
                    self.fullTreeCurrent = newfulltreecurrent

                    self.offset += Offset.SUBEXPRESSIONLENGTH
                    self.again = True
                else:  # pragma: no cover
                    errormessage = "Unkown field: %s" % (hex(self.criteria[self.offset]))
                    logging.warning(errormessage)
                    self.ignore += "Not processed: %s " % errormessage
                    logging.debug(self.criteria[self.offset:self.offset + 100])

                if not self.again:
                    break

        if self.info[Offset.LIMITBOOL] == 1:
            # Limit
            self.limit["number"] = self._iTunesUint(
                self.info[Offset.LIMITINT:Offset.LIMITINT + 4])

            self.limit["type"] = LimitMethods(
                self.info[Offset.LIMITMETHOD]).name

            # Selection
            selectionmethod = SelectionMethods(
                self.info[Offset.SELECTIONMETHOD]).name
            sign = int(self.info[Offset.SELECTIONMETHODSIGN] == 0)
            self.limit["order"] = SelectionMethodsStrings[selectionmethod][1] if isinstance(
                SelectionMethodsStrings[selectionmethod][1],
                str) else SelectionMethodsStrings[selectionmethod][1][sign]

            if len(self.output) > 0:
                self.output += '\n'
            self.output += "Limited to %d %s selected by %s" % (self.limit["number"],
                                                                self.limit["type"],
                                                                SelectionMethodsStrings[selectionmethod][0] if isinstance(
                SelectionMethodsStrings[selectionmethod][0],
                str) else SelectionMethodsStrings[selectionmethod][0][sign])

        if self.info[Offset.LIMITCHECKED] == 1:
            # Exclude unchecked items
            self.limit["onlychecked"] = True

            if len(self.output) > 0:
                self.output += '\n'
            self.output += "Exclude unchecked items"
        else:
            self.limit["onlychecked"] = False

        if self.info[Offset.LIVEUPDATE] == 0:
            # Live Update disabled
            self.limit["liveupdate"] = False

            if len(self.output) > 0:
                self.output += '\n'
            self.output += "Live updating disabled"
        else:
            self.limit["liveupdate"] = True

        t = self.limit
        t["tree"] = self.queryTree
        self.queryTree = t

        t2 = self.limit
        t2["fulltree"] = self.fullTree
        self.fullTree = t2

        self.is_parsed = True

    def ProcessStringField(self):
        end = False
        self.fieldName = StringFields(self.criteria[self.offset]).name
        self.workingOutput = self.fieldName
        self.workingQuery = "(lower(" + self.fieldName + ")"
        self.workingFull = {"field": self.fieldName, "type": "string"}

        KindEval = None

        if self.criteria[self.logicRulesOffset] == LogicRule.Contains:
            if self.criteria[self.logicSignOffset] == LogicSign.StringPositive:
                self.workingOutput += " contains "
                self.workingQuery += " LIKE '%"
                self.workingFull["operator"] = "like"
            else:
                self.workingOutput += " does not contain "
                self.workingQuery += " NOT LIKE '%"
                self.workingFull["operator"] = "not like"
            if self.criteria[self.offset] == StringFields.Kind:
                KindEval = lambda kind, query: query in kind.name
            end = True

        elif self.criteria[self.logicRulesOffset] == LogicRule.Is:
            if self.criteria[self.logicSignOffset] == LogicSign.StringPositive:
                self.workingOutput += " is "
                self.workingQuery += " = '"
                self.workingFull["operator"] = "is"
            else:
                self.workingOutput += " is not "
                self.workingQuery += " != '"
                self.workingFull["operator"] = "is not"
            if self.criteria[self.offset] == StringFields.Kind:
                KindEval = lambda kind, query: query == kind.name

        elif self.criteria[self.logicRulesOffset] == LogicRule.Starts:
            self.workingOutput += " starts with "
            self.workingQuery += " Like '"
            self.workingFull["operator"] = "starts with"
            if self.criteria[self.offset] == StringFields.Kind:
                KindEval = lambda kind, query: query not in kind.name
            end = True

        elif self.criteria[self.logicRulesOffset] == LogicRule.Ends:
            self.workingOutput += " ends with "
            self.workingQuery += " Like '%"
            self.workingFull["operator"] = "ends with"
            if self.criteria[self.offset] == StringFields.Kind:
                KindEval = lambda kind, query: kind.name.index(query) == len(kind.name) - len(query)
            end = True

        self.workingOutput += '"'
        self.content = ""
        onByte = True
        for i in range(self.stringOffset, len(self.criteria)):
            if onByte:
                if self.criteria[i] == 0 and i != len(self.criteria) - 1:
                    self.FinishStringField(end, KindEval)
                    self.offset = i + 2
                    self.again = True
                    return
                self.content += chr(self.criteria[i])
            onByte = not onByte
        self.FinishStringField(end, KindEval)
        return

    def FinishStringField(self, end, KindEval):
        self.workingOutput += self.content
        self.workingOutput += '" '
        failed = False
        if self.criteria[self.offset] == StringFields.Kind:
            self.workingQuery = ""
            for kind in FileKinds:
                if KindEval(kind, self.content):
                    if len(self.workingQuery) > 0:
                        if (len(self.query) == 0 and not self.again) or self.is_or:
                            self.workingQuery += " OR "
                        else:
                            failed = True
                            break

                    self.workingQuery += "(lower(Uri)"
                    self.workingQuery += (" LIKE '%" +
                                          kind.extension +
                                          "')") if self.criteria[self.logicSignOffset] == LogicSign.StringPositive else (" NOT LIKE '%" +
                                                                                                                         kind.extension +
                                                                                                                         "%')")
                    self.workingFull["kind_value"] = kind.extension
                    self.workingFull["kind_operator"] = "like" if self.criteria[
                        self.logicSignOffset] == LogicSign.StringPositive else "not like"
        else:
            self.workingQuery += self.content.lower()
            self.workingQuery += "%')" if end else "')"
            self.workingFull["value"] = self.content

        if len(self.ignore) > 0:
            self.ignore += self.conjunctionOutput

        if failed:
            self.ignore += self.workingOutput
        else:
            if len(self.output) > 0:
                self.output += self.conjunctionOutput
            self.output += self.workingOutput

            if len(self.query) > 0:
                self.query += self.conjunctionQuery
            self.query += self.workingQuery

            self.queryTreeCurrent.append((self.fieldName, self.workingQuery))
            self.fullTreeCurrent.append(self.workingFull)

    def ProcessIntField(self):
        self.fieldName = IntFields(self.criteria[self.offset]).name
        self.workingOutput = self.fieldName
        self.workingQuery = "(" + self.fieldName
        self.workingFull = {"field": self.fieldName, "type": "int"}

        if self.criteria[self.logicRulesOffset] == LogicRule.Is:
            number = self._iTunesUint(
                self.criteria[self.intAOffset:self.intAOffset + 4], self.criteria[self.offset] == IntFields.Rating)
            if self.criteria[self.logicSignOffset] == LogicSign.IntPositive:
                self.workingOutput += " is %d" % number
                self.workingQuery += " = %d" % number
                self.workingFull["operator"] = "is"
                self.workingFull["value"] = number
            else:
                self.workingOutput += " is not %d" % number
                self.workingQuery += " != %d" % number
                self.workingFull["operator"] = "is not"
                self.workingFull["value"] = number

        elif self.criteria[self.logicRulesOffset] == LogicRule.Greater:
            number = self._iTunesUint(
                self.criteria[self.intAOffset:self.intAOffset + 4], self.criteria[self.offset] == IntFields.Rating)
            self.workingOutput += " is greater than %d" % number
            self.workingQuery += " > %d" % number
            self.workingFull["operator"] = "greater than"
            self.workingFull["value"] = number

        elif self.criteria[self.logicRulesOffset] == LogicRule.Less:
            number = self._iTunesUint(
                self.criteria[self.intAOffset:self.intAOffset + 4], self.criteria[self.offset] == IntFields.Rating)
            self.workingOutput += " is less than %d" % number
            self.workingQuery += " < %d" % number
            self.workingFull["operator"] = "less than"
            self.workingFull["value"] = number

        elif self.criteria[self.logicRulesOffset] == LogicRule.Other:
            if self.criteria[self.logicSignOffset + 2] == 1:
                numberA = self._iTunesUint(
                    self.criteria[self.intAOffset:self.intAOffset + 4], self.criteria[self.offset] == IntFields.Rating)
                numberB = self._iTunesUint(
                    self.criteria[self.intBOffset:self.intBOffset + 4], self.criteria[self.offset] == IntFields.Rating)
                self.workingOutput += " is in the range of %d to %d" % (
                    numberA, numberB)
                self.workingQuery += " BETWEEN %d AND %d" % (numberA, numberB)
                self.workingFull["operator"] = "between"
                self.workingFull["value"] = (numberA, numberB)

            else:
                numberA = self._iTunesUint(
                    self.criteria[self.intAOffset:self.intAOffset + 4], self.criteria[self.offset] == IntFields.Rating)
                numberB = self._iTunesUint(
                    self.criteria[self.intBOffset:self.intBOffset + 4], self.criteria[self.offset] == IntFields.Rating)
                if numberA == numberB:
                    if self.criteria[self.logicSignOffset] == LogicSign.IntPositive:
                        self.workingOutput += " is %d" % numberA
                        self.workingQuery += " = %d" % numberA
                        self.workingFull["operator"] = "is"
                        self.workingFull["value"] = numberA
                    else:
                        self.workingOutput += " is not %d" % numberA
                        self.workingQuery += " != %d" % numberA
                        self.workingFull["operator"] = "is not"
                        self.workingFull["value"] = numberA
                else:  # pragma: no cover
                    errormessage = "Unkown case in ProcessIntField:LogicRule.Other: a=%d and b=%d" % (numberA, numberB)
                    logging.warning(errormessage)
                    self.ignore += " Not processed: %s " % errormessage

                    self.workingOutput += " ##UnkownCase IntField: LogicRule.Other##"
                    self.workingQuery += " ##UnkownCase IntField: LogicRule.Other##"

        self.workingQuery += ")"

        if len(self.output) > 0:
            self.output += self.conjunctionOutput

        if len(self.query) > 0:
            self.query += self.conjunctionQuery

        self.output += self.workingOutput
        self.query += self.workingQuery
        self.queryTreeCurrent.append((self.fieldName, self.workingQuery))
        self.fullTreeCurrent.append(self.workingFull)

        self.offset = self.intAOffset + Offset.INTLENGTH
        if len(self.criteria) > self.offset:
            self.again = True

    def ProcessPlaylistField(self):
        self.fieldName = PlaylistFields(self.criteria[self.offset]).name
        self.workingOutput = self.fieldName
        self.workingQuery = "(" + self.fieldName
        self.workingFull = {"field": self.fieldName, "type": "playlist"}

        if self.criteria[self.logicRulesOffset] == LogicRule.Is:
            idpart0 = self._iTunesUint(
                self.criteria[self.intAOffset - 4:self.intAOffset], self.criteria[self.offset] == IntFields.Rating)
            idpart1 = self._iTunesUint(
                self.criteria[self.intAOffset:self.intAOffset + 4], self.criteria[self.offset] == IntFields.Rating)

            if self.criteria[self.logicSignOffset] == LogicSign.IntPositive:
                self.workingOutput += " is %s%s" % (
                    hex(idpart0)[2:].upper(), hex(idpart1)[2:].upper())
                self.workingQuery += " = '%s%s'" % (
                    hex(idpart0)[2:].upper(), hex(idpart1)[2:].upper())
                self.workingFull["operator"] = "is"
                self.workingFull["value"] = "%s%s" % (
                    hex(idpart0)[2:].upper(), hex(idpart1)[2:].upper())
            else:
                self.workingOutput += " is not %s%s" % (
                    hex(idpart0)[2:].upper(), hex(idpart1)[2:].upper())
                self.workingQuery += " != '%s%s'" % (
                    hex(idpart0)[2:].upper(), hex(idpart1)[2:].upper())
                self.workingFull["operator"] = "is not"
                self.workingFull["value"] = "%s%s" % (
                    hex(idpart0)[2:].upper(), hex(idpart1)[2:].upper())

        else:  # pragma: no cover
            errormessage = "Unkown logic rule in ProcessPlaylistField: LogicRule=%d" % self.criteria[self.logicRulesOffset]
            logging.warning(errormessage)
            self.ignore += " Not processed: %s " % errormessage
            self.workingOutput += " ##UnkownCase PlaylistField:LogicRule##"
            self.workingQuery += " ##UnkownCase PlaylistField:LogicRule##"

        self.workingQuery += ")"

        if len(self.output) > 0:
            self.output += self.conjunctionOutput

        if len(self.query) > 0:
            self.query += self.conjunctionQuery

        self.output += self.workingOutput
        self.query += self.workingQuery
        self.queryTreeCurrent.append((self.fieldName, self.workingQuery))
        self.fullTreeCurrent.append(self.workingFull)

        self.offset = self.intAOffset + Offset.INTLENGTH
        if len(self.criteria) > self.offset:
            self.again = True

    def ProcessBooleanField(self):
        self.fieldName = BooleanFields(self.criteria[self.offset]).name
        self.workingOutput = self.fieldName
        self.workingQuery = "(" + self.fieldName
        self.workingFull = {"field": self.fieldName, "type": "boolean"}

        if self.criteria[self.logicRulesOffset] == LogicRule.Is:
            value = self.criteria[self.logicSignOffset] != LogicSign.IntPositive
            boolstr = "True" if value == 1 else "False"

            self.workingOutput += " is %s" % (boolstr)
            self.workingQuery += " = %d" % (value)
            self.workingFull["operator"] = "is"
            self.workingFull["value"] = value == 1

        else:  # pragma: no cover
            errormessage = "Unkown logic rule in ProcessBooleanField: LogicRule=%d" % self.criteria[self.logicRulesOffset]
            logging.warning(errormessage)
            self.ignore += " Not processed: %s " % errormessage
            self.workingOutput += " ##UnkownCase BooleanField:LogicRule##"
            self.workingQuery += " ##UnkownCase BooleanField:LogicRule##"

        self.workingQuery += ")"

        if len(self.output) > 0:
            self.output += self.conjunctionOutput

        if len(self.query) > 0:
            self.query += self.conjunctionQuery

        self.output += self.workingOutput
        self.query += self.workingQuery
        self.queryTreeCurrent.append((self.fieldName, self.workingQuery))
        self.fullTreeCurrent.append(self.workingFull)

        self.offset = self.intAOffset + Offset.INTLENGTH
        if len(self.criteria) > self.offset:
            self.again = True

    def ProcessDateField(self):
        self.fieldName = DateFields(self.criteria[self.offset]).name
        self.workingOutput = self.fieldName
        self.workingQuery = "TIMESTAMP(%s)" % self.fieldName
        self.workingFull = {"field": self.fieldName, "type": "date"}

        if self.criteria[self.logicRulesOffset] == LogicRule.Greater:
            timestamp = self._iTunesDate(
                self.criteria[self.intAOffset:self.intAOffset + 4])
            self.workingOutput += " is after %s" % self._dateString(timestamp)
            self.workingQuery += " > %d" % timestamp
            self.workingFull["operator"] = "is after"
            self.workingFull["value"] = timestamp
        elif self.criteria[self.logicRulesOffset] == LogicRule.Less:
            timestamp = self._iTunesDate(
                self.criteria[self.intAOffset:self.intAOffset + 4])
            self.workingOutput += " is before %s" % self._dateString(timestamp)
            self.workingQuery += " < %d" % timestamp
            self.workingFull["operator"] = "is before"
            self.workingFull["value"] = timestamp
        elif self.criteria[self.logicRulesOffset] == LogicRule.Other:
            if self.criteria[self.logicSignOffset + 2] == 1:
                timestampA = self._iTunesDate(
                    self.criteria[self.intAOffset:self.intAOffset + 4])
                timestampB = self._iTunesDate(
                    self.criteria[self.intBOffset:self.intBOffset + 4])
                if self.criteria[self.logicSignOffset] == LogicSign.IntPositive:
                    self.workingOutput += " is in the range of %s to %s" % (
                        self._dateString(timestampA), self._dateString(timestampB))
                    self.workingQuery += " BETWEEN %d AND %d" % (
                        timestampA, timestampB)
                    self.workingFull["operator"] = "is in the range"
                    self.workingFull["value"] = (timestampA, timestampB)
                    self.workingFull["value_date"] = (
                        self._dateString(timestampA), self._dateString(timestampB))
                else:
                    self.workingOutput += " is not in the range of %s to %s" % (
                        self._dateString(timestampA), self._dateString(timestampB))
                    self.workingQuery += " NOT BETWEEN %d AND %d" % (
                        timestampA, timestampB)
                    self.workingFull["operator"] = "is not in the range"
                    self.workingFull["value"] = (timestampA, timestampB)
                    self.workingFull["value_date"] = (
                        self._dateString(timestampA), self._dateString(timestampB))
            elif self.criteria[self.logicSignOffset + 2] == 2:
                if self.criteria[self.logicSignOffset] == LogicSign.IntPositive:
                    self.workingOutput += " is in the last "
                    self.workingQuery = "(TIMESTAMP(NOW()) - TIMESTAMP(%s)) < " % self.fieldName
                    self.workingFull["operator"] = "is in the last"
                else:
                    self.workingOutput += " is not in the last "
                    self.workingQuery = "(TIMESTAMP(NOW()) - TIMESTAMP(%s)) > " % self.fieldName
                    self.workingFull["operator"] = "is not in the last"

                t = (self._iTunesUint(bytes(
                    [255 - c for c in self.criteria[self.timeValueOffset:self.timeValueOffset + 4]])) + 1) % 4294967296
                multiple = self._iTunesUint(
                    self.criteria[self.timeMultipleOffset:self.timeMultipleOffset + 4])
                self.workingQuery += "%d" % (t * multiple)
                self.workingFull["value"] = t * multiple
                if multiple == 86400:
                    self.workingOutput += "%d days" % t
                    self.workingFull["value_date"] = "%d days" % t
                elif multiple == 604800:
                    self.workingOutput += "%d weeks" % t
                    self.workingFull["value_date"] = "%d weeks" % t
                elif multiple == 2628000:
                    self.workingOutput += "%d months" % t
                    self.workingFull["value_date"] = "%d months" % t
                else:  # pragma: no cover
                    self.workingOutput += "%d*%d seconds" % (multiple, t)
                    self.workingFull["value_date"] = "%d*%d seconds" % (
                        multiple, t)
                    errormessage = "##UnkownCase DateField: LogicRule.Other: multiple '%d' is unkown##" % multiple
                    logging.warning(errormessage)
                    self.ignore += " Not processed: %s " % errormessage

        if len(self.output) > 0:
            self.output += self.conjunctionOutput

        if len(self.query) > 0:
            self.query += self.conjunctionQuery

        self.output += self.workingOutput
        self.query += self.workingQuery
        self.queryTreeCurrent.append((self.fieldName, self.workingQuery))
        self.fullTreeCurrent.append(self.workingFull)

        self.offset = self.intAOffset + Offset.INTLENGTH
        if len(self.criteria) > self.offset:
            self.again = True

    def ProcessListField(self, fields, valueDict, listtype="list"):
        self.fieldName = fields(self.criteria[self.offset]).name
        self.workingOutput = self.fieldName
        self.workingQuery = "(" + self.fieldName
        self.workingFull = {"field": self.fieldName, "type": listtype}

        if self.criteria[self.logicRulesOffset] == LogicRule.Is:
            number = self._iTunesUint(
                self.criteria[self.intAOffset:self.intAOffset + 4], self.criteria[self.offset] == IntFields.Rating)
            if self.criteria[self.logicSignOffset] == LogicSign.IntPositive:
                self.workingOutput += " is %s" % valueDict[number]
                self.workingQuery += " = '%s'" % valueDict[number]
                self.workingFull["operator"] = "is"
                self.workingFull["value"] = valueDict[number]

            else:
                self.workingOutput += " is not %s" % valueDict[number]
                self.workingQuery += " != '%s'" % valueDict[number]
                self.workingFull["operator"] = "is not"
                self.workingFull["value"] = valueDict[number]

        elif self.criteria[self.logicRulesOffset] == LogicRule.Other:
            numberA = self._iTunesUint(
                self.criteria[self.intAOffset:self.intAOffset + 4], self.criteria[self.offset] == IntFields.Rating)
            numberB = self._iTunesUint(
                self.criteria[self.intBOffset:self.intBOffset + 4], self.criteria[self.offset] == IntFields.Rating)
            if numberA == numberB:
                if self.criteria[self.logicSignOffset] == LogicSign.IntPositive:
                    self.workingOutput += " is %s" % valueDict[numberA]
                    self.workingQuery += " = '%s'" % valueDict[numberA]
                    self.workingFull["operator"] = "is"
                    self.workingFull["value"] = valueDict[numberA]
                else:
                    self.workingOutput += " is not %s" % valueDict[numberA]
                    self.workingQuery += " != '%s'" % valueDict[numberA]
                    self.workingFull["operator"] = "is not"
                    self.workingFull["value"] = valueDict[numberA]

            else:  # pragma: no cover
                errormessage = "Unkown case in ProcessListField %s:LogicRule.Other: %d != %d" % (self.fieldName, numberA, numberB)
                logging.warning(errormessage)
                self.ignore += " Not processed: %s " % errormessage

                self.workingOutput += " ##UnkownCase ListField %s: LogicRule.Other##" % self.fieldName
                self.workingQuery += " ##UnkownCase ListField %s: LogicRule.Other##" % self.fieldName
        else:  # pragma: no cover
            errormessage = "Unkown logic rule in ProcessListField %s: LogicRule=%d" % (self.fieldName, self.criteria[self.logicRulesOffset])
            logging.warning(errormessage)
            self.ignore += " Not processed: %s " % errormessage

            self.workingOutput += " ##UnkownCase ListField %s:LogicRule##" % self.fieldName
            self.workingQuery += " ##UnkownCase ListField %s:LogicRule##" % self.fieldName

        self.workingQuery += ")"

        if len(self.output) > 0:
            self.output += self.conjunctionOutput

        if len(self.query) > 0:
            self.query += self.conjunctionQuery

        self.output += self.workingOutput
        self.query += self.workingQuery
        self.queryTreeCurrent.append((self.fieldName, self.workingQuery))
        self.fullTreeCurrent.append(self.workingFull)

        self.offset = self.intAOffset + Offset.INTLENGTH
        if len(self.criteria) > self.offset:
            self.again = True

    @staticmethod
    def _iTunesUint(bytearr, divideby=False, denominator=20):
        num = struct.unpack('>I', bytearr)[0]
        if divideby:  # For rating/stars by 20
            num = int(num / denominator)
        return num

    @classmethod
    def _iTunesDate(cls, bytearr):
        timestamp = cls._iTunesUint(bytearr)
        return timestamp + DateStartFromUnix

    @staticmethod
    def _dateString(timestamp):
        return datetime.datetime.utcfromtimestamp(
            int(timestamp)).strftime('%Y-%m-%dT%H:%M:%SZ')



import setuptools
import os
import io

with io.open("README.md", encoding="utf-8") as f:
    long_description = f.read().strip()

version = None
with io.open(os.path.join("itunessmart", "__init__.py"), encoding="utf-8") as f:
    for line in f:
        if line.strip().startswith("__version__"):
            version = line.split("=")[1].strip().replace('"', "").replace("'", "")
            break

setuptools.setup(
    name="itunessmart",
    version=version,
    license="MIT",
    author="cuzi",
    author_email="cuzi@openmail.cc",
    description="Decode iTunes Smart playlist rules. Convert iTunes smart playlists to Kodi xsp smart playlists.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cvzi/itunes_smartplaylist",
    packages=["itunessmart"],
    package_dir={"itunessmart": "itunessmart"},
    zip_safe=True,
    test_suite="nose.collector",
    tests_require=["nose"],
    classifiers=(
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Sound/Audio :: Players"
    )
)



"""
Module to convert from a parser result to a XSP playlist
"""
import logging
import hashlib
import unicodedata
import re
import html
import os.path
import collections
from typing import Callable, List, Tuple

from itunessmart.xsp_structure import *
from itunessmart.parse import SmartPlaylist

__all__ = ["createXSPFile", "createXSP", "PlaylistException", "EmptyPlaylistException"]


class PlaylistException(Exception):
    pass


class EmptyPlaylistException(PlaylistException):
    pass


def createXSPFile(directory: str, name: str, smartPlaylist: SmartPlaylist, createSubplaylists: bool = True, persistentIDMapping: dict = None, friendlyFilename: Callable[[str], str] = None) -> List[str]:
    """ Create XSP playlist file(s) from the parser result queryTree, returns a list of filenames of the generates files
    :param str directory: the output directory
    :param str name: the new name of the playlist
    :param SmartPlaylist smartPlaylist: the result of the parser
    :param bool createSubplaylists: if true subplaylists are created for nested rules/query, if false nestes rules are ommited
    :param persistentIDMapping: Optional, necessary for rules containing other playlists
    :param function friendlyFilename: Optional, function to create a filename from the playlist name: e.g. friendlyFilename = lambda name: name.strip()
    :return: list filenames
    :rtype: list
    """

    if friendlyFilename is None:
        friendlyFilename = lambda name: re.sub('[^\w\s-]', '', unicodedata.normalize('NFKD', name).encode('ascii', 'ignore').decode("utf-8")).strip()

    r = []
    for playlistname, content in createXSP(name=name, smartPlaylist=smartPlaylist, createSubplaylists=createSubplaylists, persistentIDMapping=persistentIDMapping):
        filename = friendlyFilename(playlistname) + ".xsp"
        filepath = os.path.join(directory, filename)
        with open(filepath, "wb") as f:
            f.write(content.encode("utf-8"))
        logging.info(filename)
        r.append(filename)

    return r


def createXSP(name: str, smartPlaylist: SmartPlaylist, createSubplaylists: bool = True, persistentIDMapping: dict = None, subPlaylistPrefix: str = "zzzsub_") -> List[Tuple[str, str]]:
    """ Create XSP playlist(s) from the parser result queryTree, returns a list of tuples (playlist_name, xml_content)
    :param str name: the new name of the playlist
    :param SmartPlaylist smartPlaylist: the result of the parser
    :param bool createSubplaylists: if true subplaylists are created for nested rules/query, if false nestes rules are ommited
    :param persistentIDMapping: Optional, necessary for rules containing other playlists
    :param subPlaylistPrefix: Prefix for name of sub playlists (prefix is followed by the md5 hash of the subplaylist xml)
    :return: list of tuples: (playlist_name, xml_content)
    :rtype: list
    """

    if persistentIDMapping is None:
        persistentIDMapping = {}

    queryTree = smartPlaylist.queryTree
    fulltree = queryTree["fulltree"]

    if not fulltree:
        raise EmptyPlaylistException("Playlist is empty", name)

    if "and" in fulltree:
        globalmatch = "and"
    else:
        globalmatch = "or"

    f = _minimize(_combineRules(fulltree, persistentIDMapping, createSubplaylists))

    if f:
        y = _convertRule(f, depth=0, docs=[])
        if len(y) == 3:
            # Complete doc
            globalmatch, rules, subplaylists = y
        else:
            # Only one rule
            globalmatch = "and"
            rules = y
            subplaylists = []

    else:
        raise PlaylistException("Playlist is incompatible. All of the rules are incompatible with XSP format", name)

    limit = ('    <limit>%d</limit>' % queryTree['number']) if 'number' in queryTree else ''
    order = ""
    if 'order' in queryTree:
        if queryTree['order'] == "RANDOM()":
            order = '    <order>random</order>'
        elif queryTree['order'] in xsp_sorting:
            order = '    <order direction="%s">%s</order>' % xsp_sorting[queryTree['order']]

    meta = limit + '\n' + order

    r = []
    if createSubplaylists and subplaylists:
        for sub_globalmatch, sub_rules in subplaylists:
            sub_name = subPlaylistPrefix + hashlib.md5(sub_rules.encode('utf-8')).hexdigest()
            subdocument = xml_doc.format(dec=xml_dec, name=_escapeHTML(sub_name), globalmatch=xsp_operators[sub_globalmatch], rules=sub_rules, meta="")
            rules += "\n" + xml_rule.format(field="playlist", operator="is", values=xml_value.format(value=_escapeHTML(sub_name)))
            r.append((sub_name, subdocument))

    document = xml_doc.format(dec=xml_dec, name=_escapeHTML(name), globalmatch=xsp_operators[globalmatch], rules=rules, meta=meta)

    r.append((name, document))

    return r


def _combineRules(obj, persistentIDMapping, createSubplaylists):
    """Remove incompatible rules and combine similar rules"""
    if "and" in obj or "or" in obj:
        result = []
        for operator in obj:
            t = (operator, [])
            for x in obj[operator]:
                y = _combineRules(x, persistentIDMapping, createSubplaylists)
                if y:

                    # combine with existing rule
                    combined = False
                    if operator == "or" and isinstance(y, collections.Mapping):
                        for r in t[1]:
                            if isinstance(r, collections.Mapping) and r["field"] == y["field"] and r["operator"] == y["operator"]:
                                if not isinstance(r["value"], list):
                                    r["value"] = [r["value"]]
                                r["value"].append(y["value"])
                                combined = True
                                break
                    if not combined:
                        t[1].append(y)

            if len(t[1]) > 1:
                result.append((operator, t[1]))
            elif len(t[1]) == 1 and t[1] is not None:
                result.append(t[1][0])
        return result
    else:
        if not obj["field"] in xsp_allowed_fields:
            return None
        if not obj["operator"] in xsp_allowed_operators:
            return None

        if obj["field"] == "PlaylistPersistentID":
            if obj["value"] in persistentIDMapping:
                # Replace PersistentID with playlistname
                obj["value"] = persistentIDMapping[obj["value"]]
                if not createSubplaylists:
                    try:
                        logging.warning("# Playlist dependency ignored: Depends on '%s'" % obj["value"])
                    except (UnicodeEncodeError, KeyError, TypeError) as e:
                        logging.warning("# Playlist dependency ignored: %s" % str(e))
            else:
                return None

        return obj


def _minimize(obj):
    """Remove lists with only one entry"""
    if isinstance(obj, list):
        if len(obj) == 1:
            return _minimize(obj[0])
        else:
            result = []
            for x in obj:
                y = _minimize(x)
                if y:
                    result.append(y)
            return result
    else:
        return obj

# def flat(obj, depth=0):
#     # Remove everything except first level
#     if depth > 1:
#         return
#
#     if type(obj) is tuple:
#         return (obj[0], flat(obj[1], depth+1))
#
#     if type(obj) is list:
#         result = []
#         for x in obj:
#             y = flat(x, depth)
#             if y:
#                 result.append(y)
#         return result
#
#     if type(obj) is dict:
#         return obj


def _escapeHTML(x):
    t = type(x)
    if t is str:
        return html.escape(x, quote=False)
    if t is int or t is float:
        return x

    return [_escapeHTML(i) for i in x]


def _convertRule(obj, depth, docs):
    """Create XML rules"""

    if isinstance(obj, tuple):
        if not obj[1]:
            return ""
        operator = obj[0]
        rules = []
        for x in obj[1]:
            y = _convertRule(obj=x, depth=depth+1, docs=docs)
            if y:
                rules.append(y)
        if depth > 0:
            docs.append((operator, "\n".join(rules)))
            return ""
        else:
            return operator, "\n".join(rules), docs

    elif isinstance(obj, dict):
        if "value_date" in obj:
            obj["value"] = obj["value_date"]
        if not isinstance(obj["value"], list):
            obj["value"] = [obj["value"]]
        values = []
        for value in obj["value"]:
            values.append(xml_value.format(value=_escapeHTML(value)))
        values = "\n".join(values)
        return xml_rule.format(field=xsp_fields[obj["field"]], operator=xsp_operators[obj["operator"]], values=values)

    elif isinstance(obj, list):
        rules = []
        for x in obj:
            y = _convertRule(obj=x, depth=depth, docs=docs)
            if y:
                rules.append(y)
        return "\n".join(rules)

    else:
        raise PlaylistException("Unknown obj type", repr(obj))
